class RTResult:
    def __init__(self):
        self.reset()

    def reset(self):
        self.value = None
        self.error = None
        self.func_return_value = None
        self.loop_should_continue = False
        self.loop_should_break = False

    def register(self, res):
        self.error = res.error
        self.func_return_value = res.func_return_value
        self.loop_should_continue = res.loop_should_continue
        self.loop_should_break = res.loop_should_break
        return res.value

    def success(self, value):
        self.reset()
        self.value = value
        return self

    def success_return(self, value):
        self.reset()
        self.func_return_value = value
        return self

    def success_continue(self):
        self.reset()
        self.loop_should_continue = True
        return self

    def success_break(self):
        self.reset()
        self.loop_should_break = True
        return self

    def failure(self, error):
        self.reset()
        self.error = error
        return self

    def should_return(self):
        # Note: this will allow you to continue and break outside the current function
        return (
            self.error or self.func_return_value or self.loop_should_continue or self.loop_should_break
        )



#######################################
# ARROW FUNCTION
#######################################


def show_error_loc(text, pos_start, pos_end):
    result = ''

    # Calculate indices
    idx_start = max(text.rfind('\n', 0, pos_start.idx), 0)
    idx_end = text.find('\n', idx_start + 1)
    if idx_end < 0:
        idx_end = len(text)

    # Generate each line
    line_count = pos_end.ln - pos_start.ln + 1
    for i in range(line_count):
        # Calculate line columns
        line = text[idx_start:idx_end]
        col_start = pos_start.col if i == 0 else 0
        col_end = pos_end.col if i == line_count - 1 else len(line) - 1

        # Append to result
        result += line + '\n'
        result += ' ' * col_start + '^' * (col_end - col_start)

        # Re-calculate indices
        idx_start = idx_end
        idx_end = text.find('\n', idx_start + 1)
        if idx_end < 0:
            idx_end = len(text)

    return result.replace('\t', '')

#######################################
# BASE CLASS
#######################################


class Error:
    def __init__(self, pos_start, pos_end, error_name, details):
        self.pos_start = pos_start
        self.pos_end = pos_end
        self.error_name = error_name
        self.details = details

    def as_string(self):
        result = f'{self.error_name}: {self.details}\n'
        result += f'File {self.pos_start.fn}, line {self.pos_start.ln + 1}'
        result += '\n\n' + \
            show_error_loc(self.pos_start.ftxt,
                           self.pos_start, self.pos_end)
        return result

#######################################
# SUB CLASSES
#######################################


class InvalidNumeral(Error):
    def __init__(self, pos_start, pos_end, details):
        super().__init__(pos_start, pos_end, 'Invalid Numeral', details)


class IllegalCharError(Error):
    def __init__(self, pos_start, pos_end, details):
        super().__init__(pos_start, pos_end, 'IllegalCharacterError', details)


class IndentError(Error):
    def __init__(self, pos_start, pos_end, details):
        super().__init__(pos_start, pos_end, 'IndentionError', details)


class ExpectedItemError(Error):
    def __init__(self, pos_start, pos_end, details, interactive=False):
        super().__init__(pos_start, pos_end, 'ExpectedItemError', details)
        self.interactive = interactive


class InvalidSyntaxError(Error):
    def __init__(self, pos_start, pos_end, details=''):
        super().__init__(pos_start, pos_end, 'SyntaxError', details)

#######################################
# RTERROR BASE
#######################################


class RTError(Error):
    def __init__(self, pos_start, pos_end, details, context, type_='RuntimeError'):
        super().__init__(pos_start, pos_end, type_, details)
        self.context = context

    def as_string(self):
        result = self.generate_traceback()
        result += f'{self.error_name}: {self.details}'
        result += '\n\n' + \
            show_error_loc(self.pos_start.ftxt,
                           self.pos_start, self.pos_end)
        return result

    def generate_traceback(self):
        result = ''
        pos = self.pos_start
        ctx = self.context

        while ctx:
            result = f'  File {pos.fn}, line {str(pos.ln + 1)}, in {ctx.display_name}\n' + result
            pos = ctx.parent_entry_pos
            ctx = ctx.parent

        return 'Traceback (most recent call last):\n' + result

#######################################
# RTERROR SUB CLASSES
#######################################


class TypingError(RTError):
    def __init__(self, pos_start, pos_end, details, context):
        super().__init__(pos_start, pos_end, details, context, 'TypeError')


class NamingError(RTError):
    def __init__(self, pos_start, pos_end, details, context):
        super().__init__(pos_start, pos_end, details, context, 'NameError')


class AttrError(RTError):
    def __init__(self, pos_start, pos_end, details, context):
        super().__init__(pos_start, pos_end, details, context, 'AttributeError')


class IndexingError(RTError):
    def __init__(self, pos_start, pos_end, details, context):
        super().__init__(pos_start, pos_end, details, context, 'IndexError')


class RecursingError(RTError):
    def __init__(self, pos_start, pos_end, details, context):
        super().__init__(pos_start, pos_end, details, context, 'RecursionError')


class DivisionByZeroError(RTError):
    def __init__(self, pos_start, pos_end, details, context):
        super().__init__(pos_start, pos_end, details, context, 'ZeroDivisionError')



#######################################
# IMPORTS
#######################################

from Celeratas.helper.errors import TypingError
from Celeratas.interpreter.RTResult import RTResult

#######################################
# VALUE
#######################################


class Value:
    def __init__(self):
        self.set_pos()
        self.set_context()
        self.attributes = {}

    def set_pos(self, pos_start=None, pos_end=None):
        self.pos_start = pos_start
        self.pos_end = pos_end
        return self

    def set_context(self, context=None):
        self.context = context
        return self

    def added_to(self, other):
        return None, self.illegal_operation(other)

    def subbed_by(self, other):
        return None, self.illegal_operation(other)

    def multed_by(self, other):
        return None, self.illegal_operation(other)

    def dived_by(self, other):
        return None, self.illegal_operation(other)

    def powed_by(self, other):
        return None, self.illegal_operation(other)

    def get_comparison_eq(self, other):
        return None, self.illegal_operation(other)

    def get_comparison_ne(self, other):
        return None, self.illegal_operation(other)

    def get_comparison_lt(self, other):
        return None, self.illegal_operation(other)

    def get_comparison_gt(self, other):
        return None, self.illegal_operation(other)

    def get_comparison_lte(self, other):
        return None, self.illegal_operation(other)

    def get_comparison_gte(self, other):
        return None, self.illegal_operation(other)

    def anded_by(self, other):
        return None, self.illegal_operation(other)

    def ored_by(self, other):
        return None, self.illegal_operation(other)

    def notted(self, other):
        return None, self.illegal_operation(other)

    def execute(self, args):
        return RTResult().failure(self.illegal_operation())

    def copy(self):
        raise Exception('No copy method defined')

    def is_true(self):
        return False

    def illegal_operation(self, other=None):
        if not other:
            other = self
        return TypingError(
            self.pos_start, other.pos_end,
            f'Illegal operation between {type(self).__name__} and {type(other).__name__}',
            self.context
        )



class ParseResult:
    def __init__(self):
        self.error = None
        self.node = None
        self.last_registered_advance_count = 0
        self.advance_count = 0
        self.to_reverse_count = 0

    def register_advancement(self):
        self.last_registered_advance_count = 1
        self.advance_count += 1

    def register(self, res):
        self.last_registered_advance_count = res.advance_count
        self.advance_count += res.advance_count
        if res.error:
            self.error = res.error
        return res.node

    def try_register(self, res):
        if res.error:
            self.to_reverse_count = res.advance_count
            return None
        return self.register(res)

    def success(self, node):
        self.node = node
        return self

    def failure(self, error):
        if not self.error or self.last_registered_advance_count == 0:
            self.error = error
        return self



#######################################
# IMPORTS
#######################################

import math

from .values.functions.BuiltInFunction import BuiltInFunction
from .values.Number import Number

#######################################
# CONSTANTS
#######################################

Number.null = Number(0)
Number.math_PI = Number(math.pi)

BuiltInFunction.print = BuiltInFunction("print")
BuiltInFunction.input = BuiltInFunction("input")
BuiltInFunction.clear = BuiltInFunction("clear")
BuiltInFunction.is_number = BuiltInFunction("is_number")
BuiltInFunction.is_string = BuiltInFunction("is_string")
BuiltInFunction.is_list = BuiltInFunction("is_list")
BuiltInFunction.is_function = BuiltInFunction("is_function")
BuiltInFunction.append = BuiltInFunction("append")
BuiltInFunction.pop = BuiltInFunction("pop")
BuiltInFunction.extend = BuiltInFunction("extend")
BuiltInFunction.len = BuiltInFunction("len")
BuiltInFunction.run = BuiltInFunction("run")



#######################################
# IMPORTS
#######################################

from Celeratas.helper.convert_roman import toRoman
from Celeratas.helper.errors import DivisionByZeroError

from .Bool import Bool
from .Value import Value

#######################################
# NUMBER
#######################################


class Number(Value):
    def __init__(self, value):
        super().__init__()
        self.value = value

    def added_to(self, other):
        if isinstance(other, Number) or isinstance(other, Numeral):
            return Number(self.value + other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def subbed_by(self, other):
        if isinstance(other, Number) or isinstance(other, Numeral):
            return Number(self.value - other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def multed_by(self, other):
        if isinstance(other, Number) or isinstance(other, Numeral):
            return Number(self.value * other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def dived_by(self, other):
        if isinstance(other, Number) or isinstance(other, Numeral):
            if other.value == 0:
                return None, DivisionByZeroError(
                    other.pos_start, other.pos_end,
                    'Division by zero',
                    self.context
                )

            return Number(self.value / other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def powed_by(self, other):
        if isinstance(other, Number) or isinstance(other, Numeral):
            return Number(self.value ** other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_eq(self, other):
        if isinstance(other, Number) or isinstance(other, Numeral):
            return Bool(self.value == other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_ne(self, other):
        if isinstance(other, Number) or isinstance(other, Numeral):
            return Bool(self.value != other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_lt(self, other):
        if isinstance(other, Number) or isinstance(other, Numeral):
            return Bool(self.value < other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_gt(self, other):
        if isinstance(other, Number) or isinstance(other, Numeral):
            return Bool(self.value > other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_lte(self, other):
        if isinstance(other, Number) or isinstance(other, Numeral):
            return Bool(self.value <= other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_gte(self, other):
        if isinstance(other, Number) or isinstance(other, Numeral):
            return Bool(self.value >= other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def anded_by(self, other):
        if isinstance(other, Number) or isinstance(other, Numeral):
            return Number(self.value and other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def ored_by(self, other):
        if isinstance(other, Number) or isinstance(other, Numeral):
            return Number(self.value or other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def notted(self):
        return Bool(self.value == 0).set_context(self.context), None

    def copy(self):
        copy = Number(self.value)
        copy.set_pos(self.pos_start, self.pos_end)
        copy.set_context(self.context)
        return copy

    def is_true(self):
        return self.value != 0

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return str(self.value)

#######################################
# NUMERAL
#######################################


class Numeral(Value):
    def __init__(self, value):
        super().__init__()
        self.value = value

    def added_to(self, other):
        if isinstance(other, Numeral) or isinstance(other, Number):
            return Numeral(self.value + other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def subbed_by(self, other):
        if isinstance(other, Numeral) or isinstance(other, Number):
            return Numeral(self.value - other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def multed_by(self, other):
        if isinstance(other, Numeral) or isinstance(other, Number):
            return Numeral(self.value * other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def dived_by(self, other):
        if isinstance(other, Numeral) or isinstance(other, Number):
            if other.value == 0:
                return None, DivisionByZeroError(
                    other.pos_start, other.pos_end,
                    'Division by zero',
                    self.context
                )

            return Numeral(self.value / other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def powed_by(self, other):
        if isinstance(other, Numeral) or isinstance(other, Number):
            return Numeral(self.value ** other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_eq(self, other):
        if isinstance(other, Numeral) or isinstance(other, Number):
            return Bool(self.value == other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_ne(self, other):
        if isinstance(other, Numeral) or isinstance(other, Number):
            return Bool(self.value != other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_lt(self, other):
        if isinstance(other, Numeral) or isinstance(other, Number):
            return Bool(self.value < other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_gt(self, other):
        if isinstance(other, Numeral) or isinstance(other, Number):
            return Bool(self.value > other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_lte(self, other):
        if isinstance(other, Numeral) or isinstance(other, Number):
            return Bool(self.value <= other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_gte(self, other):
        if isinstance(other, Numeral) or isinstance(other, Number):
            return Bool(self.value >= other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def anded_by(self, other):
        if isinstance(other, Numeral) or isinstance(other, Number):
            return Bool(self.value and other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def ored_by(self, other):
        if isinstance(other, Numeral) or isinstance(other, Number):
            return Bool(self.value or other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def notted(self):
        return Bool(True if self.value == 0 else False).set_context(self.context), None

    def copy(self):
        copy = Numeral(self.value)
        copy.set_pos(self.pos_start, self.pos_end)
        copy.set_context(self.context)
        return copy

    def __str__(self):
        return toRoman(round(self.value, 3)) if self.value != 0 else "nil"

    def __repr__(self):
        return toRoman(round(self.value, 3)) if self.value != 0 else "nil"



#######################################
# IMPORTS
#######################################

import Celeratas.helper.tokens as toks
from Celeratas.helper.errors import (AttrError, IndexingError, NamingError,
                                     RecursingError, RTError, TypingError)

from .RTResult import RTResult
from .values import Bool, Dict, Function, List, Number, Numeral, String

#######################################
# INTERPRETER
#######################################


class Interpreter:
    def __init__(self, recursion_depth):
        self.recursion_depth = recursion_depth

    def visit(self, node, context):
        method_name = f'visit_{type(node).__name__}'
        method = getattr(self, method_name, self.no_visit_method)
        return method(node, context)

    def no_visit_method(self, node, context):
        raise Exception(f'No visit_{type(node).__name__} method defined')

    ###################################

    def visit_NumberNode(self, node, context):
        return RTResult().success(
            Number(node.value).set_context(
                context).set_pos(node.pos_start, node.pos_end)
        )

    def visit_NumeralNode(self, node, context):
        return RTResult().success(
            Numeral(node.value).set_context(
                context).set_pos(node.pos_start, node.pos_end)
        )

    def visit_StringNode(self, node, context):
        res = RTResult()

        # Get the string
        string = ""
        for component in node.str_components:
            if isinstance(component, str):
                string += component
            else:
                result = self.visit(component, context)
                if result.error:
                    return res.failure(result.error)
                string += str(result.value)

        # Return the String Value Class
        return res.success(
            String(string).set_context(
                context).set_pos(node.pos_start, node.pos_end)
        )

    def visit_BoolNode(self, node, context):
        return RTResult().success(
            Bool(node.value == "Verus").set_context(
                context).set_pos(node.pos_start, node.pos_end)
        )

    def visit_ListNode(self, node, context):
        res = RTResult()
        elements = []

        for element_node in node.element_nodes:
            elements.append(res.register(self.visit(element_node, context)))
            if res.should_return():
                return res

        return res.success(
            List(elements).set_context(context).set_pos(
                node.pos_start, node.pos_end)
        )

    def visit_DictNode(self, node, context):
        res = RTResult()
        keypairs = {}

        for key, value in node.key_pairs.items():
            key = res.register(self.visit(key, context))

            if res.error:
                return res

            if isinstance(key, List):
                return res.failure(TypingError(
                    key.pos_start, key.pos_end,
                    "Key cannot be a list",
                    context
                ))

            value = res.register(self.visit(value, context))
            if res.error:
                return res

            keypairs[key.value] = value

        return res.success(Dict(keypairs).set_context(context).set_pos(node.pos_start, node.pos_end))

    def visit_VarAccessNode(self, node, context):
        res = RTResult()
        var_name_to_get = node.var_name_to_get
        value = context.symbol_table.get(var_name_to_get)
        idxes_to_get = node.idxes_to_get
        attrs_to_get = node.attrs_to_get

        if not value:
            return res.failure(NamingError(
                node.pos_start, node.pos_end,
                f"'{var_name_to_get}' is not defined",
                context
            ))

        value = value.copy().set_pos(node.pos_start, node.pos_end).set_context(context)

        # Check if it is not here before loop because it won't check a value that is from an index
        if idxes_to_get and not isinstance(value, List) and not isinstance(value, Dict) and not isinstance(value, String):
            return res.failure(IndexingError(
                node.pos_start, node.pos_end,
                "Can only get idx of list, dict, or string",
                context
            ))

        for idx_to_get in idxes_to_get:
            idx_to_get = res.register(self.visit(idx_to_get, context))
            idx_to_get = idx_to_get.value

            # need to get the value of the RTResult and then the Number class
            # Check if list:
            if isinstance(value, List):
                if not isinstance(idx_to_get, int):
                    return res.failure(IndexingError(
                        node.pos_start, node.pos_end,
                        'List index must be an int',
                        context
                    ))
                # If first statement is false, second will not evaluate
                if isinstance(value.elements, list) and idx_to_get < len(value.elements):
                    value.elements = value.elements[idx_to_get]
                    # If list in list:
                    if isinstance(value.elements, List):
                        value.elements = value.elements.elements
                    else:
                        value = value.elements
                else:
                    return res.failure(IndexingError(
                        node.pos_start, node.pos_end,
                        'List index out of bounds',
                        context
                    ))
            elif isinstance(value, Dict):
                found = False

                for key in value.key_pairs:
                    if key == idx_to_get:
                        value = value.key_pairs[key]
                        found = True
                        break

                if not found:
                    return res.failure(IndexingError(
                        node.pos_start, node.pos_end,
                        'Dict index out of bounds',
                        context
                    ))
            elif isinstance(value, String):
                if not isinstance(idx_to_get, int):
                    return res.failure(IndexingError(
                        node.pos_start, node.pos_end,
                        'String index must be an int',
                        context
                    ))
                if idx_to_get < len(value.value):
                    value.value = value.value[idx_to_get]
                else:
                    return res.failure(IndexingError(
                        node.pos_start, node.pos_end,
                        'String index out of bounds',
                        context
                    ))
            else:
                return res.failure(TypingError(
                    node.pos_start, node.pos_end,
                    'Index out of bounds',
                    context
                ))

        for attr_to_get in attrs_to_get:
            value_attr = value.attributes.get(attr_to_get, None)
            if value_attr is None:
                return res.failure(AttrError(
                    node.pos_start, node.pos_end,
                    f"'{var_name_to_get}' does not have the attribute '{attr_to_get}'",
                    context
                ))

            value = value_attr
        return res.success(value)

    def visit_VarAssignNode(self, node, context):
        res = RTResult()
        assign_type = node.assign_type

        if len(node.vars_to_set) != len(node.values_to_set):
            return res.failure(TypingError(
                node.pos_start, node.pos_end,
                'Must have the same number of variables and values',
                context
            ))

        for var, value in zip(node.vars_to_set, node.values_to_set):
            value = res.register(self.visit(value, context))
            var_name = var[0]
            idxes_to_change = var[1]

            if res.should_return():
                return res

            if idxes_to_change:
                if var_name not in context.symbol_table.symbols:
                    return res.failure(NamingError(
                        node.pos_start, node.pos_end,
                        f"'{var_name}' is not defined",
                        context
                    ))

                var_to_change = context.symbol_table.get(var_name)

                element_to_change = var_to_change

                for for_idx, idx_to_change in enumerate(idxes_to_change):
                    if isinstance(element_to_change, List):
                        if idx_to_change.value > len(element_to_change.elements) - 1:
                            return res.failure(IndexingError(
                                node.pos_start, node.pos_end,
                                'List index out of bounds',
                                context
                            ))
                        if for_idx == len(idxes_to_change) - 1:
                            element_to_change.elements[idx_to_change.value] = value
                        else:
                            element_to_change = element_to_change.elements[idx_to_change.value]
                    elif isinstance(element_to_change, Dict) or (hasattr(element_to_change, "elements") and isinstance(element_to_change.elements, Dict)):
                        if for_idx == len(idxes_to_change) - 1:
                            element_to_change.key_pairs[idx_to_change.value] = value
                        else:
                            element_to_change = element_to_change.key_pairs[idx_to_change.value]
                    else:
                        return res.failure(IndexingError(
                            node.pos_start, node.pos_end,
                            f'Variable must be a list or dict not \'{type(element_to_change).__name__}\' in order to set a specific index',
                            context
                        ))

                value = var_to_change

            if assign_type.type != "EQ":
                error = None
                old_value = context.symbol_table.get(var_name)

                if old_value is None:
                    return res.failure(NamingError(
                        node.pos_start, node.pos_end,
                        f"'{var_name}' is not defined",
                        context
                    ))

                if assign_type.type == "PLUS_EQ":
                    value, error = old_value.added_to(value)
                elif assign_type.type == "MIN_EQ":
                    value, error = old_value.subbed_by(value)
                elif assign_type.type == "MUL_EQ":
                    value, error = old_value.multed_by(value)
                elif assign_type.type == "DIV_EQ":
                    value, error = old_value.dived_by(value)

                if error:
                    return res.failure(error)

            context.symbol_table.set(var_name, value)

        return res.success(None)

    def visit_BinOpNode(self, node, context):
        res = RTResult()
        left = res.register(self.visit(node.left_node, context))
        if res.should_return():
            return res
        right = res.register(self.visit(node.right_node, context))
        if res.should_return():
            return res

        if node.op_tok.type == toks.TT_PLUS:
            result, error = left.added_to(right)
        elif node.op_tok.type == toks.TT_MINUS:
            result, error = left.subbed_by(right)
        elif node.op_tok.type == toks.TT_MUL:
            result, error = left.multed_by(right)
        elif node.op_tok.type == toks.TT_DIV:
            result, error = left.dived_by(right)
        elif node.op_tok.type == toks.TT_POW:
            result, error = left.powed_by(right)
        elif node.op_tok.type == toks.TT_EE:
            result, error = left.get_comparison_eq(right)
        elif node.op_tok.type == toks.TT_NE:
            result, error = left.get_comparison_ne(right)
        elif node.op_tok.type == toks.TT_LT:
            result, error = left.get_comparison_lt(right)
        elif node.op_tok.type == toks.TT_GT:
            result, error = left.get_comparison_gt(right)
        elif node.op_tok.type == toks.TT_LTE:
            result, error = left.get_comparison_lte(right)
        elif node.op_tok.type == toks.TT_GTE:
            result, error = left.get_comparison_gte(right)
        elif node.op_tok.matches(toks.TT_KEYWORD, 'et'):
            result, error = left.anded_by(right)
        elif node.op_tok.matches(toks.TT_KEYWORD, 'aut'):
            result, error = left.ored_by(right)

        if error:
            return res.failure(error)
        else:
            return res.success(result.set_pos(node.pos_start, node.pos_end))

    def visit_UnaryOpNode(self, node, context):
        res = RTResult()
        number = res.register(self.visit(node.node, context))
        if res.should_return():
            return res

        error = None

        if node.op_tok.type == toks.TT_MINUS:
            number, error = number.multed_by(Number(-1))
        elif node.op_tok.matches(toks.TT_KEYWORD, 'non'):
            number, error = number.notted()

        if error:
            return res.failure(error)
        else:
            return res.success(number.set_pos(node.pos_start, node.pos_end))

    def visit_IfNode(self, node, context):
        res = RTResult()

        for condition, expr, should_return_null in node.cases:
            condition_value = res.register(self.visit(condition, context))
            if res.should_return():
                return res

            if condition_value is None:
                return res.failure(RTError(
                    node.pos_start, node.pos_end,
                    'Conditional can not be evaluated',
                    context
                ))

            if condition_value.is_true():
                expr_value = res.register(self.visit(expr, context))
                if res.should_return():
                    return res
                return res.success(None if should_return_null else expr_value)

        if node.else_case:
            expr, should_return_null = node.else_case
            expr_value = res.register(self.visit(expr, context))
            if res.should_return():
                return res
            return res.success(None if should_return_null else expr_value)

        return res.success(None)

    def visit_TryNode(self, node, context):
        res = RTResult()
        try_body = self.visit(node.try_body, context)

        if node.except_name and node.except_name.value not in ["Exception", "TypeError", "NameError", "IndexError", "ZeroDivisionError"]:
            return res.failure(NamingError(
                node.except_name.pos_start, node.except_name.pos_end,
                'Exception type not supported',
                context
            ))

        should_except = not node.except_name or node.except_name.value == "Exception" or try_body.error.error_name == node.except_name.value
        if try_body.error and node.except_body and should_except:
            if node.except_as:
                context.symbol_table.set(
                    node.except_as.value, String(try_body.error.details))

            res.register(self.visit(node.except_body, context))
            if res.error:
                return res

            if node.except_as:
                context.symbol_table.remove(node.except_as.value)
        elif try_body.error is None:
            try_body = res.register(try_body)
            if res.error:
                return res

        return res.success(None if node.should_return_null else try_body)

    def visit_ForNode(self, node, context):
        res = RTResult()
        elements = []

        start_value = res.register(self.visit(node.start_value_node, context))
        if res.should_return():
            return res

        end_value = res.register(self.visit(node.end_value_node, context))
        if res.should_return():
            return res

        if node.step_value_node:
            step_value = res.register(
                self.visit(node.step_value_node, context))
            if res.should_return():
                return res
        else:
            step_value = Number(1)

        if start_value is None:
            return res.failure(RTError(
                node.start_value_node.pos_start, node.start_value_node.pos_end,
                'Expression does not have a value',
                context
            ))

        i = start_value.value

        if step_value.value >= 0:
            def condition():
                return i < end_value.value
        else:
            def condition():
                return i > end_value.value

        while condition():
            context.symbol_table.set(node.var_name, Number(i))
            i += step_value.value

            value = res.register(self.visit(node.body_node, context))
            if res.should_return() and res.loop_should_continue and res.loop_should_break:
                return res

            if res.loop_should_continue:
                continue

            if res.loop_should_break:
                break

            elements.append(value)

        context.symbol_table.remove(node.var_name)
        return res.success(
            None if node.should_return_null else
            List(elements).set_context(context).set_pos(
                node.pos_start, node.pos_end)
        )

    def visit_WhileNode(self, node, context):
        res = RTResult()
        elements = []

        while True:
            condition_value = res.register(
                self.visit(node.condition_node, context))
            if res.should_return():
                return res

            if condition_value is None:
                return res.failure(RTError(
                    node.condition_node.pos_start, node.condition_node.pos_end,
                    'Conditional can not be evaluated',
                    context
                ))

            if not condition_value.is_true():
                break

            value = res.register(self.visit(node.body_node, context))
            if res.should_return() and res.loop_should_continue and res.loop_should_break:
                return res

            if res.loop_should_continue:
                continue

            if res.loop_should_break:
                break

            elements.append(value)

        return res.success(
            None if node.should_return_null else
            List(elements).set_context(context).set_pos(
                node.pos_start, node.pos_end)
        )

    def visit_FuncDefNode(self, node, context):
        res = RTResult()

        func_name = node.func_name if node.func_name else None
        body_node = node.body_node
        args = []
        for arg_name, arg_value in node.args:
            if arg_value is None:
                args.append((arg_name, arg_value))
            else:
                arg_value = res.register(self.visit(arg_value, context))
                if res.should_return():
                    return res
                args.append((arg_name, arg_value))

        func_value = Function(func_name, body_node, args, node.should_auto_return).set_context(
            context).set_pos(node.pos_start, node.pos_end)

        if node.func_name:
            context.symbol_table.set(func_name, func_value)

        return res.success(func_value)

    def visit_CallNode(self, node, context):

        if self.recursion_depth > 100:
            return RTResult().failure(RecursingError(
                node.pos_start, node.pos_end,
                'Recursion depth exceeded',
                context
            ))

        res = RTResult()
        args = {}

        value_to_call = res.register(self.visit(node.node_to_call, context))
        if res.should_return():
            return res
        value_to_call = value_to_call.copy().set_pos(node.pos_start, node.pos_end)

        for arg_key, arg_value in node.arg_nodes.items():
            arg_value = res.register(self.visit(arg_value, context))
            if res.should_return():
                return res

            args[arg_key] = arg_value

        return_value = res.register(value_to_call.execute(args, self.recursion_depth))
        if res.should_return():
            return res
        if return_value is None:
            return res.success(None)

        return_value = return_value.copy().set_pos(
            node.pos_start, node.pos_end).set_context(context)
        return res.success(return_value)

    def visit_ReturnNode(self, node, context):
        res = RTResult()

        if node.node_to_return:
            value = res.register(self.visit(node.node_to_return, context))
            if res.should_return():
                return res
        else:
            value = None

        return res.success_return(value)

    def visit_ContinueNode(self, node, context):
        return RTResult().success_continue()

    def visit_BreakNode(self, node, context):
        return RTResult().success_break()



#####################################
# IMPORTS
#####################################

from Celeratas.helper.errors import TypingError
from Celeratas.interpreter.Context import Context
from Celeratas.interpreter.RTResult import RTResult
from Celeratas.interpreter.SymbolTable import SymbolTable
from Celeratas.interpreter.values.Value import Value

#####################################
# BASE FUNCTION
#####################################


class BaseFunction(Value):
    def __init__(self, name):
        super().__init__()
        self.name = name or "<anonymous>"

    def generate_new_context(self):
        new_context = Context(self.name, self.context, self.pos_start)
        new_context.symbol_table = SymbolTable(new_context.parent.symbol_table)
        return new_context

    def check_args(self, func_args, input_args):

        res = RTResult()

        if len(input_args) > len(func_args):
            return res.failure(TypingError(
                self.pos_start, self.pos_end,
                f"{len(input_args) - len(func_args)} too many args passed into {self.name}()",
                self.context
            ))

        pos_args = [i for i in func_args if i[1] is None]

        if len(input_args) < len(pos_args):
            return res.failure(TypingError(
                self.pos_start, self.pos_end,
                f"{len(pos_args) - len(input_args)} too few args passed into {self}()",
                self.context
            ))

        return res.success(None)

    def populate_args(self, func_args, input_args, exec_ctx):

        for arg_idx, arg in enumerate(func_args):
            arg_name = arg[0]
            arg_value = arg[1]

            if arg_name in input_args:
                input_arg_value = input_args[arg_name]
                input_arg_value.set_context(exec_ctx)
                exec_ctx.symbol_table.set(arg_name, input_arg_value)
            elif arg_idx in input_args:
                input_arg_value = input_args[arg_idx]
                input_arg_value.set_context(exec_ctx)
                exec_ctx.symbol_table.set(arg_name, input_arg_value)
            else:
                arg_value.set_context(exec_ctx)
                exec_ctx.symbol_table.set(arg_name, arg_value)

    def check_and_populate_args(self, arg_names, args, exec_ctx):
        res = RTResult()
        res.register(self.check_args(arg_names, args))
        if res.should_return():
            return res
        self.populate_args(arg_names, args, exec_ctx)
        return res.success(None)



#######################################
# IMPORTS
#######################################

import re

#######################################
# LOOKUP
#######################################

lookup = [
    (1000, 'M'),
    (900, 'CM'),
    (500, 'D'),
    (400, 'CD'),
    (100, 'C'),
    (90, 'XC'),
    (50, 'L'),
    (40, 'XL'),
    (10, 'X'),
    (9, 'IX'),
    (5, 'V'),
    (4, 'IV'),
    (1, 'I'),
]

#######################################
# Functions
#######################################


def toRoman(number):
    if isinstance(number, float):
        if number.is_integer():
            number = int(number)
        else:
            before_dot, after_dot = str(number).split(".")
            before_dot, after_dot = toRoman(
                int(before_dot)), toRoman(int(after_dot))
            return f"{before_dot}.{after_dot}"
    if number < 0:
        number = toRoman(number * -1)
        return f"-{number}"

    res = ''
    for (n, roman) in lookup:
        (d, number) = divmod(number, n)
        res += roman * d
    return res


def toNum(numeral):
    valid = bool(re.search(
        r"^-?M{0,3}(CM|CD|D?C{0,3})(XC|XL|L?X{0,3})(IX|IV|V?I{0,3})(\..M{0,3}(CM|CD|D?C{0,3})(XC|XL|L?X{0,3})(IX|IV|V?I{0,3}))?$", numeral))
    if not valid:
        return None

    if "-" in numeral:
        numeral = toNum(numeral[1:])
        return numeral * -1

    if "." in numeral:
        before_dot, after_dot = numeral.split(".")
        before_dot, after_dot = toNum(before_dot), toNum(after_dot)
        if before_dot is None or after_dot is None:
            return None
        return float(f"{before_dot}.{after_dot}")

    res = 0
    numeral = numeral.replace("IV", "IIII").replace("IX", "VIIII")
    numeral = numeral.replace("XL", "XXXX").replace("XC", "LXXXX")
    numeral = numeral.replace("CD", "CCCC").replace("CM", "DCCCC")
    for char in numeral:
        for n in lookup:
            if char == n[1]:
                res += n[0]
    return res



#######################################
# IMPORTS
#######################################

import Celeratas.helper.tokens as toks
from Celeratas.helper.errors import (ExpectedItemError, IndentError,
                                     InvalidSyntaxError)

from .nodes import (BinOpNode, BoolNode, BreakNode, CallNode, ContinueNode,
                    DictNode, ForNode, FuncDefNode, IfNode, ListNode,
                    NumberNode, NumeralNode, ReturnNode, StringNode, TryNode,
                    UnaryOpNode, VarAccessNode, VarAssignNode, WhileNode)
from .ParseResult import ParseResult

#######################################
# PARSER
#######################################


class Parser:
    def __init__(self, tokens):
        self.tokens = tokens
        self.tok_idx = -1
        self.indent_count = 0
        self.advance()

    def advance(self):
        self.tok_idx += 1
        self.update_current_tok()
        return self.current_tok

    def reverse(self, amount=1):
        self.tok_idx -= amount
        self.update_current_tok()
        return self.current_tok

    def update_current_tok(self):
        if self.tok_idx >= 0 and self.tok_idx < len(self.tokens):
            self.current_tok = self.tokens[self.tok_idx]

    def parse(self):
        res = self.statements(in_loop=False, in_func=False)

        if not res.error and self.current_tok.type != toks.TT_EOF:
            return res.failure(InvalidSyntaxError(
                self.current_tok.pos_start, self.current_tok.pos_end,
                "Token cannot appear after previous tokens"
            ))
        return res

    def check_indent_amount(self):
        res = ParseResult()
        tab_count = 0

        # Check Tab count from tokens
        while True:
            if self.current_tok.type == toks.TT_TAB:
                res.register_advancement()
                self.advance()
                tab_count += 1
            else:
                break

        # Check if greater than expected tab count
        if tab_count > self.indent_count:
            return None, res.failure(IndentError(
                self.current_tok.pos_start, self.current_tok.pos_end,
                "Incorrect number of tabs!"))
        return tab_count, None

    ###################################

    def statements(self, in_loop, in_func):
        res = ParseResult()
        statements = []
        pos_start = self.current_tok.pos_start.copy()

        while self.current_tok.type == toks.TT_NEWLINE:
            res.register_advancement()
            self.advance()

        tab_amount, error = self.check_indent_amount()
        if error:
            res.register(error)
            return res

        if tab_amount < self.indent_count:
            return res.failure(ExpectedItemError(
                self.current_tok.pos_start, self.current_tok.pos_end,
                "Expected Tab"))

        statement = res.register(self.statement(in_loop, in_func))

        if res.error:
            return res

        statements.append(statement)
        more_statements = True

        while True:
            newline_count = 0
            while self.current_tok.type == toks.TT_NEWLINE:
                res.register_advancement()
                self.advance()
                newline_count += 1

            tab_amount, error = self.check_indent_amount()

            if error:
                res.register(error)
                return res

            if tab_amount < self.indent_count:
                more_statements = False

            if not more_statements:
                break

            statement = res.try_register(self.statement(in_loop, in_func))

            if not statement:
                self.reverse(res.to_reverse_count)
                more_statements = False
                continue

            statements.append(statement)

        return res.success(ListNode(
            statements,
            pos_start,
            self.current_tok.pos_end.copy()
        ))

    def statement(self, in_loop, in_func):
        res = ParseResult()
        pos_start = self.current_tok.pos_start.copy()

        if self.current_tok.matches(toks.TT_KEYWORD, 'redi'):
            if not in_func:
                return res.failure(ExpectedItemError(
                    self.current_tok.pos_start, self.current_tok.pos_end,
                    "Return can only be used in functions"
                ))

            res.register_advancement()
            self.advance()

            expr = res.try_register(self.expr(in_loop, in_func))
            if not expr:
                self.reverse(res.to_reverse_count)
            return res.success(ReturnNode(expr, pos_start, self.current_tok.pos_start.copy()))

        if self.current_tok.matches(toks.TT_KEYWORD, 'continua'):
            if not in_loop:
                return res.failure(ExpectedItemError(
                    self.current_tok.pos_start, self.current_tok.pos_end,
                    "Continue can only be used in loops"
                ))

            res.register_advancement()
            self.advance()
            return res.success(ContinueNode(pos_start, self.current_tok.pos_start.copy()))

        if self.current_tok.matches(toks.TT_KEYWORD, 'confringe'):
            if not in_loop:
                return res.failure(ExpectedItemError(
                    self.current_tok.pos_start, self.current_tok.pos_end,
                    "Break can only be used in loops"
                ))

            res.register_advancement()
            self.advance()
            return res.success(BreakNode(pos_start, self.current_tok.pos_start.copy()))

        expr = res.register(self.expr(in_loop, in_func))
        if res.error:
            return res.failure(ExpectedItemError(
                self.current_tok.pos_start, self.current_tok.pos_end,
                "Expected expression"
            ))
        return res.success(expr)

    def expr(self, in_loop, in_func):
        res = ParseResult()

        if self.current_tok.type == toks.TT_IDENTIFIER:
            vars_to_set = []
            pos_start = self.current_tok.pos_start.copy()

            while True:
                var_name = self.current_tok.value
                idxes_to_change = []

                res.register_advancement()
                self.advance()

                while self.current_tok.type == toks.TT_LSQUARE:
                    res.register_advancement()
                    self.advance()

                    idxes_to_change.append(res.register(self.bin_op(in_loop, in_func,
                                                                    self.comp_expr, ((toks.TT_KEYWORD, 'et'), (toks.TT_KEYWORD, 'aut')))))

                    if res.error:
                        return res

                    res.register_advancement()
                    self.advance()

                vars_to_set.append([var_name, idxes_to_change])

                if not self.current_tok.type == toks.TT_COMMA:
                    break

                res.register_advancement()
                self.advance()

                if self.current_tok.type != toks.TT_IDENTIFIER:
                    return res.failure(ExpectedItemError(
                        self.current_tok.pos_start, self.current_tok.pos_end,
                        "Expected identifier"
                    ))

            if self.current_tok.type not in [toks.TT_EQ, toks.TT_PLUS_EQ, toks.TT_MIN_EQ, toks.TT_MUL_EQ, toks.TT_DIV_EQ]:
                self.reverse(res.advance_count)
            else:
                assign_type = self.current_tok

                res.register_advancement()
                self.advance()

                values = []

                while True:
                    values.append(res.register(
                        self.bin_op(in_loop, in_func,
                                    self.comp_expr, ((toks.TT_KEYWORD, 'et'), (toks.TT_KEYWORD, 'aut')))))

                    if res.error:
                        return res

                    if not self.current_tok.type == toks.TT_COMMA:
                        break

                    res.register_advancement()
                    self.advance()

                return res.success(VarAssignNode(vars_to_set, values, assign_type, pos_start, pos_end=values[-1].pos_end))

        node = res.register(self.bin_op(in_loop, in_func,
                                        self.comp_expr, ((toks.TT_KEYWORD, 'et'), (toks.TT_KEYWORD, 'aut'))))

        if res.error:
            return res.failure(ExpectedItemError(
                self.current_tok.pos_start, self.current_tok.pos_end,
                "Expected expression"
            ))

        return res.success(node)

    def comp_expr(self, in_loop, in_func):
        res = ParseResult()

        if self.current_tok.matches(toks.TT_KEYWORD, 'non'):
            op_tok = self.current_tok
            res.register_advancement()
            self.advance()

            node = res.register(self.comp_expr())
            if res.error:
                return res
            return res.success(UnaryOpNode(op_tok, node, op_tok.pos_start, node.pos_end))

        node = res.register(self.bin_op(in_loop, in_func,
                                        self.arith_expr, (toks.TT_EE, toks.TT_NE, toks.TT_LT, toks.TT_GT, toks.TT_LTE, toks.TT_GTE)))

        if res.error:
            return res.failure(ExpectedItemError(
                self.current_tok.pos_start, self.current_tok.pos_end,
                "Expected expression"
            ))

        return res.success(node)

    def arith_expr(self, in_loop, in_func):
        return self.bin_op(in_loop, in_func, self.term, (toks.TT_PLUS, toks.TT_MINUS))

    def term(self, in_loop, in_func):
        return self.bin_op(in_loop, in_func, self.factor, (toks.TT_MUL, toks.TT_DIV))

    def factor(self, in_loop, in_func):
        res = ParseResult()
        tok = self.current_tok

        if tok.type in (toks.TT_PLUS, toks.TT_MINUS):
            res.register_advancement()
            self.advance()
            factor = res.register(self.factor(in_loop, in_func))
            if res.error:
                return res
            return res.success(UnaryOpNode(tok, factor, tok.pos_start, factor.pos_end))

        return self.power(in_loop, in_func)

    def power(self, in_loop, in_func):
        return self.bin_op(in_loop, in_func, self.call, (toks.TT_POW, ), self.factor)

    def call(self, in_loop, in_func):
        res = ParseResult()
        pos_start = self.current_tok.pos_start.copy()
        atom = res.register(self.atom(in_loop, in_func))
        if res.error:
            return res

        if self.current_tok.type == toks.TT_LPAREN:
            res.register_advancement()
            self.advance()
            arg_nodes = {}

            if self.current_tok.type == toks.TT_RPAREN:
                res.register_advancement()
                self.advance()
            else:
                keyword_arg = False
                new_keyword = False
                arg_idx = 0

                while True:
                    if self.current_tok.type == toks.TT_IDENTIFIER:
                        arg_name = self.current_tok.value
                        for prev_name in arg_nodes:
                            if arg_name == prev_name:
                                res.failure(InvalidSyntaxError(
                                    self.current_tok.pos_start, self.current_tok.pos_end,
                                    f"Duplicate argument '{arg_name}'"
                                ))
                                return res
                        self.advance()

                        if self.current_tok.type != toks.TT_EQ:
                            self.reverse(1)
                        else:
                            res.register_advancement()
                            res.register_advancement()
                            self.advance()

                            arg_value = res.register(self.bin_op(in_loop, in_func,
                                                                 self.comp_expr, ((toks.TT_KEYWORD, 'et'), (toks.TT_KEYWORD, 'aut'))))
                            if res.error:
                                return res

                            arg_nodes[arg_name] = arg_value
                            keyword_arg = True
                            new_keyword = True

                    if not new_keyword:
                        if keyword_arg:
                            return res.failure(InvalidSyntaxError(
                                self.current_tok.pos_start, self.current_tok.pos_end,
                                "Positional argument cannot follow keyword argument"
                            ))

                        arg_nodes[arg_idx] = res.register(self.bin_op(in_loop, in_func,
                                                                      self.comp_expr, ((toks.TT_KEYWORD, 'et'), (toks.TT_KEYWORD, 'aut'))))
                        if res.error:
                            return res

                    arg_idx += 1

                    if self.current_tok.type != toks.TT_COMMA:
                        break
                    res.register_advancement()
                    self.advance()

                if self.current_tok.type != toks.TT_RPAREN:
                    return res.failure(ExpectedItemError(
                        self.current_tok.pos_start, self.current_tok.pos_end,
                        "Expected ',' or ')'"
                    ))

                res.register_advancement()
                self.advance()
            return res.success(CallNode(atom, arg_nodes, pos_start, atom.pos_end))
        return res.success(atom)

    def atom(self, in_loop, in_func):
        res = ParseResult()
        tok = self.current_tok

        if tok.type == toks.TT_NUMERAL:
            res.register_advancement()
            self.advance()
            return res.success(NumeralNode(tok.value, tok.pos_start, tok.pos_end))

        if tok.type in (toks.TT_INT, toks.TT_FLOAT):
            res.register_advancement()
            self.advance()
            return res.success(NumberNode(tok.value, tok.pos_start, tok.pos_end))

        elif tok.type == toks.TT_STRING:
            str_components = tok.value
            for val in str_components:
                if isinstance(val, list):
                    parser = Parser(val)
                    ast = parser.expr(in_loop, in_func)
                    if ast.error:
                        return res.failure(ast.error)
                    str_components[str_components.index(val)] = ast.node

            res.register_advancement()
            self.advance()
            return res.success(StringNode(str_components, tok.pos_start, tok.pos_end))

        elif tok.matches(toks.TT_KEYWORD, 'Verus') or tok.matches(toks.TT_KEYWORD, 'Falsus'):
            res.register_advancement()
            self.advance()
            return res.success(BoolNode(tok.value, tok.pos_start, tok.pos_end))

        elif tok.type == toks.TT_IDENTIFIER:
            var_access = res.register(self.var_access_expr(in_loop, in_func))
            if res.error:
                return res
            return res.success(var_access)

        elif tok.matches(toks.TT_KEYWORD, 'opus') or tok.type == toks.TT_LPAREN:
            func_def = self.func_def(in_loop, in_func)
            # If LPAREN, then func will check if it's a anonymous function and return the number of advances if it is not, the code below will then reverse the operation
            if isinstance(func_def, int):
                self.reverse(func_def)
            elif res.error:
                return res
            else:
                func_def = res.register(func_def)
                return res.success(func_def)

        # For Order of operations
        if tok.type == toks.TT_LPAREN:
            res.register_advancement()
            self.advance()
            expr = res.register(self.expr(in_loop, in_func))
            if res.error:
                return res
            if self.current_tok.type == toks.TT_RPAREN:
                res.register_advancement()
                self.advance()
                return res.success(expr)
            else:
                return res.failure(ExpectedItemError(
                    self.current_tok.pos_start, self.current_tok.pos_end,
                    "Expected ')'"
                ))

        elif tok.type == toks.TT_LSQUARE:
            list_expr = res.register(self.list_expr(in_loop, in_func))
            if res.error:
                return res
            return res.success(list_expr)
        elif tok.type == toks.TT_LBRACE:
            dict_expr = res.register(self.dict_expr(in_loop, in_func))
            if res.error:
                return res
            return res.success(dict_expr)
        elif tok.matches(toks.TT_KEYWORD, 'si'):
            if_expr = res.register(self.if_expr(in_loop, in_func))
            if res.error:
                return res
            return res.success(if_expr)
        elif tok.matches(toks.TT_KEYWORD, 'tempta'):
            try_expr = res.register(self.try_expr(in_loop, in_func))
            if res.error:
                return res
            return res.success(try_expr)
        elif tok.matches(toks.TT_KEYWORD, 'pro'):
            for_expr = res.register(self.for_expr(in_func))
            if res.error:
                return res
            return res.success(for_expr)

        elif tok.matches(toks.TT_KEYWORD, 'dum'):
            while_expr = res.register(self.while_expr(in_func))
            if res.error:
                return res
            return res.success(while_expr)

        return res.failure(ExpectedItemError(
            tok.pos_start, tok.pos_end,
            "Expected expression"
        ))

    def var_access_expr(self, in_loop, in_func):
        res = ParseResult()

        var_name = self.current_tok.value
        pos_start = self.current_tok.pos_start
        pos_end = self.current_tok.pos_end
        idxes_to_get = []
        attrs_to_get = []

        res.register_advancement()
        self.advance()

        if self.current_tok.type == toks.TT_LSQUARE:
            while self.current_tok.type == toks.TT_LSQUARE:
                res.register_advancement()
                self.advance()

                idxes_to_get.append(res.register(self.bin_op(in_loop, in_func,
                                                             self.comp_expr, ((toks.TT_KEYWORD, 'et'), (toks.TT_KEYWORD, 'aut')))))

                if res.error:
                    return res

                pos_end = self.current_tok.pos_end
                res.register_advancement()
                self.advance()

        while self.current_tok.type == toks.TT_DOT:
            res.register_advancement()
            self.advance()

            if self.current_tok.type != toks.TT_IDENTIFIER:
                return res.failure(ExpectedItemError(
                    self.current_tok.pos_start, self.current_tok.pos_end,
                    "Expected Identifier"
                ))

            attrs_to_get.append(self.current_tok.value)
            pos_end = self.current_tok.pos_end

            res.register_advancement()
            self.advance()

        return res.success(VarAccessNode(var_name, idxes_to_get, attrs_to_get, pos_start, pos_end))

    def list_expr(self, in_loop, in_func):
        res = ParseResult()
        element_nodes = []
        pos_start = self.current_tok.pos_start.copy()

        res.register_advancement()
        self.advance()

        if self.current_tok.type == toks.TT_RSQUARE:
            res.register_advancement()
            self.advance()
        else:
            element_nodes.append(res.register(
                self.bin_op(in_loop, in_func,
                            self.comp_expr, ((toks.TT_KEYWORD, 'et'), (toks.TT_KEYWORD, 'aut')))))
            if res.error:
                return res.failure(ExpectedItemError(
                    self.current_tok.pos_start, self.current_tok.pos_end,
                    "Expected ']' or expression"
                ))

            while self.current_tok.type == toks.TT_COMMA:
                res.register_advancement()
                self.advance()

                element_nodes.append(res.register(self.expr(in_loop, in_func)))
                if res.error:
                    return res

            if self.current_tok.type != toks.TT_RSQUARE:
                return res.failure(ExpectedItemError(
                    self.current_tok.pos_start, self.current_tok.pos_end,
                    "Expected ',' or ']'"
                ))

            res.register_advancement()
            self.advance()

        return res.success(ListNode(
            element_nodes,
            pos_start,
            self.current_tok.pos_end.copy()
        ))

    def dict_expr(self, in_func, in_loop):
        res = ParseResult()
        key_pairs = {}
        pos_start = self.current_tok.pos_start.copy()

        res.register_advancement()
        self.advance()

        if self.current_tok.type == toks.TT_RBRACE:
            res.register_advancement()
            self.advance()
        else:
            while True:
                key = res.register(self.bin_op(in_loop, in_func,
                                               self.comp_expr, ((toks.TT_KEYWORD, 'et'), (toks.TT_KEYWORD, 'aut'))))
                if res.error:
                    return res

                if self.current_tok.type != toks.TT_COLON:
                    return res.failure(ExpectedItemError(
                        self.current_tok.pos_start, self.current_tok.pos_end,
                        "Expected ':'"
                    ))

                res.register_advancement()
                self.advance()

                value = res.register(self.bin_op(in_loop, in_func,
                                                 self.comp_expr, ((toks.TT_KEYWORD, 'et'), (toks.TT_KEYWORD, 'aut'))))
                if res.error:
                    return res

                key_pairs[key] = value

                if self.current_tok.type != toks.TT_COMMA:
                    break

                res.register_advancement()
                self.advance()

            res.register_advancement()
            self.advance()

        return res.success(DictNode(key_pairs, pos_start, self.current_tok.pos_end.copy()))

    def if_expr(self, in_loop, in_func):
        res = ParseResult()
        all_cases = res.register(self.if_expr_cases('si', in_loop, in_func))
        if res.error:
            return res
        cases, else_case = all_cases
        return res.success(IfNode(cases, else_case, cases[0][0].pos_start, (else_case or cases[len(cases) - 1])[0].pos_end))

    def if_expr_b(self, in_loop, in_func):
        return self.if_expr_cases('alioquinsi', in_loop, in_func)

    def if_expr_c(self, in_loop, in_func):
        res = ParseResult()
        else_case = None

        if self.current_tok.matches(toks.TT_KEYWORD, 'alioquin'):
            res.register_advancement()
            self.advance()

            if self.current_tok.type == toks.TT_COLON:
                res.register_advancement()
                self.advance()
            else:
                return res.failure(ExpectedItemError(
                    self.current_tok.pos_start, self.current_tok.pos_end,
                    "Expected ':'"
                ))

            if self.current_tok.type == toks.TT_NEWLINE:
                res.register_advancement()
                self.advance()

                self.indent_count += 1
                statements = res.register(self.statements(in_loop, in_func))
                self.indent_count -= 1

                if res.error:
                    return res
                else_case = (statements, True)

            else:
                expr = res.register(self.statement(in_loop, in_func))
                if res.error:
                    return res
                else_case = (expr, False)

        return res.success(else_case)

    def if_expr_b_or_c(self, in_loop, in_func):
        res = ParseResult()
        cases, else_case = [], None

        if self.current_tok.matches(toks.TT_KEYWORD, 'alioquinsi'):
            all_cases = res.register(self.if_expr_b(in_loop, in_func))
            if res.error:
                return res
            cases, else_case = all_cases
        else:
            else_case = res.register(self.if_expr_c(in_loop, in_func))
            if res.error:
                return res

        return res.success((cases, else_case))

    def if_expr_cases(self, case_keyword, in_loop, in_func):
        res = ParseResult()
        cases = []
        else_case = None

        if not self.current_tok.matches(toks.TT_KEYWORD, case_keyword):
            return res.failure(ExpectedItemError(
                self.current_tok.pos_start, self.current_tok.pos_end,
                f"Expected '{case_keyword}'"
            ))

        res.register_advancement()
        self.advance()

        condition = res.register(self.expr(in_loop, in_func))
        if res.error:
            return res

        if not self.current_tok.type == toks.TT_COLON:
            return res.failure(ExpectedItemError(
                self.current_tok.pos_start, self.current_tok.pos_end,
                "Expected ':'"
            ))

        res.register_advancement()
        self.advance()

        if self.current_tok.type == toks.TT_EOF:
            return res.failure(ExpectedItemError(
                self.current_tok.pos_start, self.current_tok.pos_end,
                "Expected statement", interactive=True
            ))

        if self.current_tok.type == toks.TT_NEWLINE:
            res.register_advancement()
            self.advance()

            self.indent_count += 1
            statements = res.register(self.statements(in_loop, in_func))
            self.indent_count -= 1

            if res.error:
                return res
            cases.append((condition, statements, True))

            all_cases = res.register(self.if_expr_b_or_c(in_loop, in_func))
            if res.error:
                return res
            new_cases, else_case = all_cases
            cases.extend(new_cases)
        else:
            expr = res.register(self.statement(in_loop, in_func))
            if res.error:
                return res
            cases.append((condition, expr, False))

            all_cases = res.register(self.if_expr_b_or_c(in_loop, in_func))
            if res.error:
                return res
            new_cases, else_case = all_cases
            cases.extend(new_cases)

        return res.success((cases, else_case))

    def try_expr(self, in_loop, in_func):
        res = ParseResult()
        pos_start = self.current_tok.pos_start.copy()

        res.register_advancement()
        self.advance()

        if self.current_tok == toks.TT_COLON:
            return res.failure(ExpectedItemError(
                self.current_tok.pos_start, self.current_tok.pos_end,
                "Expected ':'"
            ))

        res.register_advancement()
        self.advance()

        if self.current_tok.type == toks.TT_EOF:
            return res.failure(ExpectedItemError(
                self.current_tok.pos_start, self.current_tok.pos_end,
                "Expected statement", interactive=True
            ))

        if self.current_tok.type == toks.TT_NEWLINE:
            res.register_advancement()
            self.advance()

            self.indent_count += 1
            try_body = res.register(self.statements(in_loop, in_func))
            self.indent_count -= 1
            if res.error:
                return res

            should_return_null = True

            if self.current_tok.matches(toks.TT_KEYWORD, "praeter"):
                self.advance()
                res.register_advancement()

                if self.current_tok.type == toks.TT_IDENTIFIER:
                    except_name = self.current_tok
                    res.register_advancement()
                    self.advance()
                else:
                    except_name = None

                if self.current_tok.matches(toks.TT_KEYWORD, "tam"):
                    res.register_advancement()
                    self.advance()

                    if self.current_tok.type != toks.TT_IDENTIFIER:
                        return res.failure(ExpectedItemError(
                            self.current_tok.pos_start, self.current_tok.pos_end,
                            "Expected identifier"
                        ))

                    except_as = self.current_tok
                    res.register_advancement()
                    self.advance()
                else:
                    except_as = None

                if self.current_tok == toks.TT_COLON:
                    return res.failure(ExpectedItemError(
                        self.current_tok.pos_start, self.current_tok.pos_end,
                        "Expected ':'"
                    ))

                self.advance()
                res.register_advancement()

                self.indent_count += 1
                except_body = res.register(self.statements(in_loop, in_func))
                self.indent_count -= 1

                if res.error:
                    return res
            else:
                except_body = None
                except_name = None
                except_as = None
        else:
            try_body = res.register(self.expr(in_loop, in_func))
            if res.error:
                return res

            except_body = None
            except_name = None
            except_as = None
            should_return_null = False

        return res.success(TryNode(try_body, except_body, except_name, except_as, should_return_null, pos_start, except_body.pos_end if except_body else try_body.pos_end))

    def for_expr(self, in_func):
        res = ParseResult()
        pos_start = self.current_tok.pos_start.copy()

        res.register_advancement()
        self.advance()

        if self.current_tok.type != toks.TT_IDENTIFIER:
            return res.failure(ExpectedItemError(
                self.current_tok.pos_start, self.current_tok.pos_end,
                "Expected identifier"
            ))

        var_name = self.current_tok.value
        res.register_advancement()
        self.advance()

        if self.current_tok.type != toks.TT_EQ:
            return res.failure(ExpectedItemError(
                self.current_tok.pos_start, self.current_tok.pos_end,
                "Expected '='"
            ))

        res.register_advancement()
        self.advance()

        start_value = res.register(self.bin_op(True, in_func,
                                               self.comp_expr, ((toks.TT_KEYWORD, 'et'), (toks.TT_KEYWORD, 'aut'))))
        if res.error:
            return res

        if not self.current_tok.matches(toks.TT_KEYWORD, 'ad'):
            return res.failure(ExpectedItemError(
                self.current_tok.pos_start, self.current_tok.pos_end,
                "Expected 'ad'"
            ))

        res.register_advancement()
        self.advance()

        end_value = res.register(self.bin_op(True, in_func,
                                             self.comp_expr, ((toks.TT_KEYWORD, 'et'), (toks.TT_KEYWORD, 'aut'))))
        if res.error:
            return res

        if self.current_tok.matches(toks.TT_KEYWORD, 'gradus'):
            res.register_advancement()
            self.advance()

            step_value = res.register(self.bin_op(True, in_func,
                                                  self.comp_expr, ((toks.TT_KEYWORD, 'et'), (toks.TT_KEYWORD, 'aut'))))
            if res.error:
                return res
        else:
            step_value = None

        if not self.current_tok.type == toks.TT_COLON:
            return res.failure(ExpectedItemError(
                self.current_tok.pos_start, self.current_tok.pos_end,
                "Expected ':'"
            ))

        res.register_advancement()
        self.advance()

        if self.current_tok.type == toks.TT_EOF:
            return res.failure(ExpectedItemError(
                self.current_tok.pos_start, self.current_tok.pos_end,
                "Expected statement", interactive=True
            ))

        if self.current_tok.type == toks.TT_NEWLINE:
            res.register_advancement()
            self.advance()

            self.indent_count += 1
            body = res.register(self.statements(in_loop=True, in_func=in_func))
            self.indent_count -= 1

            if res.error:
                return res

            return res.success(ForNode(var_name, start_value, end_value, step_value, body, True, pos_start, body.pos_end))

        body = res.register(self.statement(in_loop=True, in_func=in_func))
        if res.error:
            return res

        return res.success(ForNode(var_name, start_value, end_value, step_value, body, False, pos_start, body.pos_end))

    def while_expr(self, in_func):
        res = ParseResult()
        pos_start = self.current_tok.pos_start.copy()

        res.register_advancement()
        self.advance()

        condition = res.register(self.expr(in_loop=True, in_func=in_func))
        if res.error:
            return res

        if not self.current_tok.type == toks.TT_COLON:
            return res.failure(ExpectedItemError(
                self.current_tok.pos_start, self.current_tok.pos_end,
                "Expected ':'"
            ))

        res.register_advancement()
        self.advance()

        if self.current_tok.type == toks.TT_EOF:
            return res.failure(ExpectedItemError(
                self.current_tok.pos_start, self.current_tok.pos_end,
                "Expected statement", interactive=True
            ))

        if self.current_tok.type == toks.TT_NEWLINE:
            res.register_advancement()
            self.advance()

            self.indent_count += 1
            body = res.register(self.statements(in_loop=True, in_func=in_func))
            self.indent_count -= 1

            if res.error:
                return res

            return res.success(WhileNode(condition, body, True, pos_start, body.pos_end))

        body = res.register(self.statement(in_loop=True, in_func=in_func))
        if res.error:
            return res

        return res.success(WhileNode(condition, body, False, pos_start, body.pos_end))

    def _register_args(self):
        res = ParseResult()

        res.register_advancement()
        self.advance()

        args = []
        key_word_arg = False

        if self.current_tok.type == toks.TT_IDENTIFIER:
            while True:
                if self.current_tok.type != toks.TT_IDENTIFIER:
                    return res.failure(ExpectedItemError(
                        self.current_tok.pos_start, self.current_tok.pos_end,
                        "Expected identifier"
                    ))

                arg_name = self.current_tok.value
                for prev_arg_name, _ in args:
                    if arg_name == prev_arg_name:
                        return res.failure(InvalidSyntaxError(
                            self.current_tok.pos_start, self.current_tok.pos_end,
                            f"Duplicate argument '{arg_name}'"
                        ))

                res.register_advancement()
                self.advance()

                if self.current_tok.type == toks.TT_EQ:
                    res.register_advancement()
                    self.advance()
                    arg_value = res.register(self.bin_op(False, False, self.comp_expr, ((toks.TT_KEYWORD, 'et'), (toks.TT_KEYWORD, 'aut'))))
                    if res.error:
                        return res

                    args.append((arg_name, arg_value))
                    key_word_arg = True
                else:
                    if key_word_arg:
                        return res.failure(InvalidSyntaxError(
                            self.current_tok.pos_start, self.current_tok.pos_end,
                            "Positional argument cannot follow keyword argument"
                        ))

                    args.append((arg_name, None))

                if not self.current_tok.type == toks.TT_COMMA:
                    break
                res.register_advancement()
                self.advance()

            if self.current_tok.type != toks.TT_RPAREN:
                return res.failure(ExpectedItemError(
                    self.current_tok.pos_start, self.current_tok.pos_end,
                    "Expected ',' or ')'"
                ))

        else:
            if self.current_tok.type != toks.TT_RPAREN:
                return res.failure(ExpectedItemError(
                    self.current_tok.pos_start, self.current_tok.pos_end,
                    "Expected identifier or ')'"
                ))

        res.register_advancement()
        self.advance()
        return res.success(args)

    def func_def(self, in_loop, in_func):
        # Need helper function because anynomous and full-fledged functions are handled differently
        # Call this function when current token is '('

        res = ParseResult()
        pos_start = self.current_tok.pos_start.copy()

        # Anonymous function
        if self.current_tok.type == toks.TT_LPAREN:
            arg_name_toks = res.register(self._register_args())

            # Need to divide advance count by 2 for weird inner function reasons
            if res.error:
                return res.advance_count

            if self.current_tok.type != toks.TT_ARROW:
                return res.advance_count

            res.register_advancement()
            self.advance()

            body = res.register(self.expr(in_loop, in_func))
            if res.error:
                return res

            return res.success(FuncDefNode(
                None,
                arg_name_toks,
                body,
                True,
                pos_start,
                body.pos_end
            ))

        # Advance here past opus keyword

        res.register_advancement()
        self.advance()

        if self.current_tok.type != toks.TT_IDENTIFIER:
            return res.failure(ExpectedItemError(
                self.current_tok.pos_start, self.current_tok.pos_end,
                "Expected identifier"
            ))

        var_name_tok = self.current_tok.value
        res.register_advancement()
        self.advance()

        if self.current_tok.type != toks.TT_LPAREN:
            return res.failure(ExpectedItemError(
                self.current_tok.pos_start, self.current_tok.pos_end,
                "Expected '('"
            ))

        arg_name_toks = res.register(self._register_args())
        if res.error:
            return res

        if self.current_tok.type != toks.TT_COLON:
            return res.failure(ExpectedItemError(
                self.current_tok.pos_start, self.current_tok.pos_end,
                "Expected ':'"
            ))

        res.register_advancement()
        self.advance()

        if self.current_tok.type == toks.TT_EOF:
            return res.failure(ExpectedItemError(
                self.current_tok.pos_start, self.current_tok.pos_end,
                "Expected statement", interactive=True
            ))

        if self.current_tok.type != toks.TT_NEWLINE:
            return res.failure(ExpectedItemError(
                self.current_tok.pos_start, self.current_tok.pos_end,
                "Expected a newline"
            ))

        res.register_advancement()
        self.advance()

        self.indent_count += 1
        body = res.register(self.statements(in_loop=False, in_func=True))
        self.indent_count -= 1

        if res.error:
            return res

        res.register_advancement()
        self.advance()

        return res.success(FuncDefNode(var_name_tok, arg_name_toks, body, False, pos_start, body.pos_end))

    ###################################

    def bin_op(self, in_loop, in_func, func_a, ops, func_b=None):
        if func_b is None:
            func_b = func_a

        res = ParseResult()
        left = res.register(func_a(in_loop, in_func))
        if res.error:
            return res

        while self.current_tok.type in ops or (self.current_tok.type, self.current_tok.value) in ops:
            op_tok = self.current_tok
            res.register_advancement()
            self.advance()
            right = res.register(func_b(in_loop, in_func))
            if res.error:
                return res
            left = BinOpNode(left, op_tok, right, left.pos_start, right.pos_end)

        return res.success(left)



class Context:
    def __init__(self, display_name, parent=None, parent_entry_pos=None):
        self.display_name = display_name
        self.parent = parent
        self.parent_entry_pos = parent_entry_pos
        self.symbol_table = None



class Position:
    def __init__(self, idx, ln, col, fn, ftxt):
        self.idx = idx
        self.ln = ln
        self.col = col
        self.fn = fn
        self.ftxt = ftxt

    def advance(self, current_char=None):
        self.idx += 1
        self.col += 1

        if current_char == '\n':
            self.ln += 1
            self.col = 0

        return self

    def copy(self):
        return Position(self.idx, self.ln, self.col, self.fn, self.ftxt)



#######################################
# IMPORTS
#######################################

import sys
from datetime import datetime

import Celeratas.interpreter.constants as constants

from .interpreter.Context import Context
from .interpreter.Interpreter import Interpreter
from .interpreter.SymbolTable import SymbolTable
from .lexer.Lexer import Lexer
from .parser.Parser import Parser

#######################################
# GLOBAL SYMBOL TABLE
#######################################

global_symbol_table = SymbolTable()
global_symbol_table.set("nil", constants.Number.null)
global_symbol_table.set("pi", constants.Number.math_PI)
global_symbol_table.set("scribe", constants.BuiltInFunction.print)
global_symbol_table.set("initus", constants.BuiltInFunction.input)
global_symbol_table.set("purgo", constants.BuiltInFunction.clear)
global_symbol_table.set("est_numerus", constants.BuiltInFunction.is_number)
global_symbol_table.set("est_filum", constants.BuiltInFunction.is_string)
global_symbol_table.set("est_album", constants.BuiltInFunction.is_list)
global_symbol_table.set("est_opus", constants.BuiltInFunction.is_function)
global_symbol_table.set("adde", constants.BuiltInFunction.append)
global_symbol_table.set("remove", constants.BuiltInFunction.pop)
global_symbol_table.set("extende", constants.BuiltInFunction.extend)
global_symbol_table.set("longitudo", constants.BuiltInFunction.len)
global_symbol_table.set("curre", constants.BuiltInFunction.run)
#######################################
# RUN SCRIPT FUNCTION
#######################################

# Function Out of class because curre() needs to access it


def run_script(fn, text):
    # Generate tokens
    lexer = Lexer(fn, text)
    tokens, error = lexer.make_tokens()
    if error:
        return None, error

    # Generate AST
    parser = Parser(tokens)
    ast = parser.parse()
    if ast.error:
        return None, ast.error

    # Run program
    interpreter = Interpreter(recursion_depth=0)
    context = Context('<program>')
    context.symbol_table = global_symbol_table
    result = interpreter.visit(ast.node, context)

    return result.value, result.error


class Shell:
    def __init__(self):
        self.start()

    #######################################
    # DATE
    #######################################

    def get_time(self):
        time = datetime.now().strftime("%b %d %y, %H:%M:%S")
        time = time.replace("Jan", "Unus Martius").replace(
            "Feb", "Duo Martius")
        time = time.replace("Mar", "Martius").replace("Apr", "Aprilis")
        time = time.replace("May", "Maius").replace("Jun", "Junius")
        time = time.replace("Jul", "Quintilis").replace("Aug", "Sextilis")
        return time
        # Don't need to replace Sep, Oct, Nov, or Dec bc they are the same in latin

    #######################################
    # HELP FUNCTION
    #######################################

    def help_menu(self):
        while True:
            print(
                "\nWhat do you need help with?\n\n1: Keywords\n2: Syntax\n3: Built-In Functions\n4: Leave\n")
            while True:
                help_num = input()
                if help_num in ["1", "2", "3", "4"]:
                    break
                else:
                    print("\nPlease only type 1, 2, 3, or 4\n")

            if help_num == "1":
                print('\nVerus -> True\nFalsus -> False\net -> and\naut -> or\nnon -> not\nsi -> if')
                print('alioquinsi -> elif\nalioquin -> else\ntempta -> try\npraeter -> except\tam -> as\nnpro -> for')
                print('ad -> to\ngradus -> step\ndum -> while\nopus -> task\nfinis -> end\nredi -> return')
                print('continua -> continue\nconfringe -> break')
            elif help_num == "2":
                with open("src/helper/grammar.txt", "r") as grammar:
                    print(f"\n{grammar.read()}")
            elif help_num == "3":
                print('\nscribe(value_to_print) -> print\ninitus() -> input\npurgo() -> clear\nest_numerus(value_to_check) -> is_number')
                print('est_filum(value_to_check) -> is_string\nest_album(value_to_check) -> is_list\nest_opus(value_to_check) -> is_function')
                print('adde(value_to_add) -> append\nremove(value_to_remove) -> pop\nextende(list_to_extend) -> extend\nlongitudo(value_to_check) -> length')
                print('curre(file_to_run) -> run')
            else:
                break

    #######################################
    # GET RESULT FUNCTION
    #######################################

    def get_result(self, fn, script, interactive):
        result, error = run_script(fn, script)

        if hasattr(error, "interactive") and error.interactive:
            return True
        elif error:
            print(error.as_string())
        elif result:
            result.elements = [x for x in result.elements if repr(x) != "None"]
            if len(result.elements) == 0:
                pass
            elif len(result.elements) == 1:
                print(repr(result.elements[0]))
            else:
                print(repr(result))
        return False

    #######################################
    # RUN FILE FROM CLI ARGS OR TAKE INPUT
    #######################################

    def start(self):
        try:
            if len(sys.argv) > 1:
                try:
                    fn = sys.argv[1]
                    with open(fn, "r") as f:
                        script = f.read()
                    self.get_result(fn, script, interactive=False)
                except FileNotFoundError:
                    print(f"Can't open file {fn}: No such file")
                except UnicodeDecodeError:
                    print(f"Can't open file {fn}: Invalid file format")
            else:
                time = self.get_time()
                print(
                    f"Celeritas versio unum (defalta, {time})\nScribe 'auxilium' auxilio")
                more_statements = False
                statements_txt = ""

                while True:
                    text = input("... " if more_statements else ">>> ")

                    if text == "auxilium" and not more_statements:
                        self.help_menu()
                        continue

                    if text.strip() == "":
                        if more_statements:
                            self.get_result(
                                "<stdin>", statements_txt, interactive=False)
                            more_statements = False
                            continue
                        else:
                            continue

                    if more_statements:
                        statements_txt += f"\n{text.rjust(len(text) + 4)}"
                    else:
                        error = self.get_result(
                            "<stdin>", text, interactive=False)
                        # Error is true when program throws an interactive error
                        if error:
                            more_statements = True
                            statements_txt = text

        except KeyboardInterrupt:
            print("\r", end="")
            print("Keyboard Interrupt")

# For entry point:


def start():
    Shell()



class Token:
    def __init__(self, type_, value=None, pos_start=None, pos_end=None):
        self.type = type_
        self.value = value

        if pos_start:
            self.pos_start = pos_start.copy()
            self.pos_end = pos_start.copy()
            self.pos_end.advance()

        if pos_end:
            self.pos_end = pos_end.copy()

    def matches(self, type_, value):
        return self.type == type_ and self.value == value

    def __repr__(self):
        if self.value:
            return f'{self.type}:{self.value}'
        return f'{self.type}'



#######################################
# CONSTANTS
#######################################


TT_INT = 'INT'
TT_FLOAT = 'FLOAT'
TT_STRING = 'STRING'
TT_NUMERAL = 'NUMERAL'
TT_IDENTIFIER = 'IDENTIFIER'
TT_DOT = 'DOT'
TT_KEYWORD = 'KEYWORD'
TT_PLUS = 'PLUS'
TT_MINUS = 'MINUS'
TT_MUL = 'MUL'
TT_DIV = 'DIV'
TT_POW = 'POW'
TT_EQ = 'EQ'
TT_PLUS_EQ = 'PLUS_EQ'
TT_MIN_EQ = 'MIN_EQ'
TT_MUL_EQ = "MUL_EQ"
TT_DIV_EQ = "DIV_EQ"
TT_LPAREN = 'LPAREN'
TT_RPAREN = 'RPAREN'
TT_LSQUARE = 'LSQUARE'
TT_RSQUARE = 'RSQUARE'
TT_LBRACE = 'LBRACE'
TT_RBRACE = 'RBRACE'
TT_EE = 'EE'
TT_NE = 'NE'
TT_LT = 'LT'
TT_GT = 'GT'
TT_LTE = 'LTE'
TT_GTE = 'GTE'
TT_COLON = 'COLON'
TT_COMMA = 'COMMA'
TT_ARROW = 'ARROW'
TT_TAB = 'TAB'
TT_NEWLINE = 'NEWLINE'
TT_EOF = 'EOF'

KEYWORDS = [
    'Verus',
    'Falsus',
    'et',
    'aut',
    'non',
    'si',
    'alioquinsi',
    'alioquin',
    'tempta',
    'praeter',
    'tam',
    'pro',
    'ad',
    'gradus',
    'dum',
    'opus',
    'redi',
    'continua',
    'confringe',
]



#######################################
# IMPORTS
#######################################

from .Bool import Bool
from .Number import Number
from .Value import Value

#######################################
# STRING
#######################################


class String(Value):
    def __init__(self, value):
        super().__init__()
        self.value = value

        self.attributes = {"length": Number(len(value))}

    def added_to(self, other):
        if isinstance(other, String):
            return String(self.value + other.value, self.length + other.length).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def multed_by(self, other):
        if isinstance(other, Number):
            return String(self.value * other.value, self.length * other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_eq(self, other):
        if isinstance(other, String):
            return Bool(self.value == other.value), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_ne(self, other):
        if isinstance(other, String):
            return Bool(self.value != other.value), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_lt(self, other):
        if isinstance(other, String):
            return Bool(self.value < other.value), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_gt(self, other):
        if isinstance(other, String):
            return Bool(self.value > other.value), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_lte(self, other):
        if isinstance(other, String):
            return Bool(self.value <= other.value), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_gte(self, other):
        if isinstance(other, String):
            return Bool(self.value >= other.value), None
        else:
            return None, Value.illegal_operation(self, other)

    def anded_by(self, other):
        if isinstance(other, String):
            return String(self.value and other.value), None
        else:
            return None, Value.illegal_operation(self, other)

    def ored_by(self, other):
        if isinstance(other, String):
            return String(self.value or other.value), None
        else:
            return None, Value.illegal_operation(self, other)

    def notted(self):
        return Bool(len(self.value) == 0), None

    def is_true(self):
        return len(self.value) > 0

    def copy(self):
        copy = String(self.value)
        copy.set_pos(self.pos_start, self.pos_end)
        copy.set_context(self.context)
        return copy

    def __str__(self):
        return self.value

    def __repr__(self):
        return f'"{self.value}"'



#######################################
# IMPORTS
#######################################

import Celeratas.helper.tokens as toks
import Celeratas.lexer.constants as constants
from Celeratas.helper.convert_roman import toNum
from Celeratas.helper.errors import (ExpectedItemError, IllegalCharError,
                                     IndentError, InvalidNumeral)

from .Position import Position
from .Token import Token

#######################################
# LEXER
#######################################


class Lexer:
    def __init__(self, fn, text):
        self.fn = fn
        self.text = text
        self.pos = Position(-1, 0, -1, fn, text)
        self.current_char = None
        self.tab_style = ""
        self.advance()

    def advance(self):
        self.pos.advance(self.current_char)
        self.current_char = self.text[self.pos.idx] if self.pos.idx < len(
            self.text) else None

    def make_tokens(self):
        tokens = []
        self.start_of_statement = True
        # Start of statements is before character other than space/tab - After this is false, tabs and spaces are ignored by make_spaces() method
        # Grammatical char is set to True is anything other than space/tab is read - This var updates start of statements at the end of while loop
        while self.current_char:
            grammatical_char = True
            if self.current_char in ' \t':
                new_toks, error = self.make_space(tokens)
                if error:
                    return [], error
                tokens = new_toks
                grammatical_char = False
            elif self.current_char == '#':
                self.skip_comment()
                self.start_of_statement = True
                grammatical_char = False
            elif self.current_char in ';\n':
                tokens.append(Token(toks.TT_NEWLINE, pos_start=self.pos))
                self.advance()
                self.start_of_statement = True
                grammatical_char = False
            elif self.current_char in constants.ROMAN_NUMERAL_CHARS:
                token, error = self.make_numeral()
                if error:
                    return [], error
                tokens.append(token)
            elif self.current_char in constants.LETTERS + "_":
                token, error = self.make_identifier()
                if error:
                    return [], error
                tokens.append(token)
            elif self.current_char == ".":
                tokens.append(Token(toks.TT_DOT, pos_start=self.pos))
                self.advance()
            elif self.current_char in constants.DIGITS:
                tokens.append(self.make_number())
            elif self.current_char == '"' or self.current_char == "'":
                # Need the first clause incase the code starts with a quote
                if len(tokens) > 0 and tokens[-1].matches(toks.TT_IDENTIFIER, 'f'):
                    tokens.pop()
                    token, error = self.make_string(fstring=True)
                else:
                    token, error = self.make_string(fstring=False)

                if error:
                    return [], error
                tokens.append(token)
            elif self.current_char == '+':
                tokens.append(self.make_plus())
            elif self.current_char == '-':
                tokens.append(self.make_minus())
            elif self.current_char == '*':
                tokens.append(self.make_mul())
                self.advance()
            elif self.current_char == '/':
                tokens.append(self.make_div())
                self.advance()
            elif self.current_char == '^':
                tokens.append(Token(toks.TT_POW, pos_start=self.pos))
                self.advance()
            elif self.current_char == '(':
                tokens.append(Token(toks.TT_LPAREN, pos_start=self.pos))
                self.advance()
            elif self.current_char == ')':
                tokens.append(Token(toks.TT_RPAREN, pos_start=self.pos))
                self.advance()
            elif self.current_char == '[':
                tokens.append(Token(toks.TT_LSQUARE, pos_start=self.pos))
                self.advance()
            elif self.current_char == ']':
                tokens.append(Token(toks.TT_RSQUARE, pos_start=self.pos))
                self.advance()
            elif self.current_char == '{':
                tokens.append(Token(toks.TT_LBRACE, pos_start=self.pos))
                self.advance()
            elif self.current_char == '}':
                tokens.append(Token(toks.TT_RBRACE, pos_start=self.pos))
                self.advance()
            elif self.current_char == '!':
                token, error = self.make_not_equals()
                if error:
                    return [], error
                tokens.append(token)
            elif self.current_char == '=':
                tokens.append(self.make_equals())
            elif self.current_char == '<':
                tokens.append(self.make_less_than())
            elif self.current_char == '>':
                tokens.append(self.make_greater_than())
            elif self.current_char == ':':
                tokens.append(Token(toks.TT_COLON, pos_start=self.pos))
                self.advance()
            elif self.current_char == ',':
                tokens.append(Token(toks.TT_COMMA, pos_start=self.pos))
                self.advance()
            else:
                pos_start = self.pos.copy()
                char = self.current_char
                self.advance()
                return [], IllegalCharError(pos_start, self.pos, "'" + char + "'")

            # When Grammatical Token is read, tabs/spaces from this point on will be ignored
            if self.start_of_statement and grammatical_char:
                self.start_of_statement = False

        tokens.append(Token(toks.TT_EOF, pos_start=self.pos))
        return tokens, None

    def make_space(self, tokens):
        pos_start = self.pos.copy()
        if self.start_of_statement:
            if self.current_char == ' ':
                # Check tab style
                if self.tab_style not in ["", "space"]:
                    return [], IndentError(pos_start, self.pos, "Inconsistent indentation")
                # Check if next 3 chars are also space
                for _ in range(3):
                    self.advance()
                    if self.current_char == " ":
                        continue
                    # Throw error if not
                    return None, IndentError(pos_start, self.pos, "Improper indentation")

                self.advance()
                self.tab_style = "space"
                tokens.append(
                    Token(toks.TT_TAB, pos_start=pos_start, pos_end=self.pos))
            elif self.current_char == '\t':
                # Check tab style
                if self.tab_style not in ["", "tab"]:
                    return [], IndentError(pos_start, self.pos, "Inconsistent indentation")

                self.advance()
                self.tab_style = "tab"
                tokens.append(Token(toks.TT_TAB, pos_start=self.pos))
        else:
            self.advance()
        return tokens, None

    def make_numeral(self):
        numeral_str = ''
        dot_count = 0
        pos_start = self.pos.copy()

        while self.current_char and self.current_char in constants.ROMAN_NUMERAL_CHARS + '.':
            if self.current_char == '.':
                if dot_count == 1:
                    break
                dot_count += 1
            numeral_str += self.current_char
            self.advance()
        # Check if identifier
        if self.current_char and self.current_char != " ":
            # Return the identifier
            identifier, error = self.make_identifier(numeral_str)
            # Need this extra code because the position of the pos start of the identifier will be behind the pos start of the numeral
            if error:
                return "", error
            return Token(identifier.type, identifier.value, pos_start, identifier.pos_end), None
        # Convert String to Int or Float
        numeral_final = toNum(numeral_str)

        if numeral_final is None:
            return None, InvalidNumeral(pos_start, self.pos, f"{numeral_str} is not a valid numeral")

        return Token(toks.TT_NUMERAL, numeral_final, pos_start, self.pos), None

    def make_number(self):
        num_str = ''
        dot_count = 0
        pos_start = self.pos.copy()

        while self.current_char and self.current_char in constants.DIGITS + '.':
            if self.current_char == '.':
                if dot_count == 1:
                    break
                dot_count += 1
            num_str += self.current_char
            self.advance()

        if dot_count == 0:
            return Token(toks.TT_INT, int(num_str), pos_start, self.pos)
        else:
            return Token(toks.TT_FLOAT, float(num_str), pos_start, self.pos)

    def make_string(self, fstring):
        pos_start = self.pos.copy()
        escape_character = False
        # Can be either " or '
        string_char = self.current_char
        self.advance()

        escape_characters = {
            'n': '\n',
            't': '\t'
        }

        str_components = []
        cur_str = ""

        if not self.current_char:
            return None, IllegalCharError(pos_start, self.pos, "Unexpected end to string")

        while self.current_char != string_char:
            if escape_character:
                # If the escape character is not in the lookup, it simply adds the character
                cur_str += escape_characters.get(self.current_char,
                                                 self.current_char)
                escape_character = False
            elif self.current_char == '\\':
                # This affect the next iteration
                escape_character = True
            elif self.current_char == "\n":
                return None, IllegalCharError(pos_start, self.pos, "Unexpected end to string")
            elif fstring and self.current_char == "{":
                # Append cur_str to str_components and reset cur_str
                if cur_str != "":
                    str_components.append(cur_str)
                    cur_str = ""

                # Put the text between braces into the string to_lex
                to_lex = ""
                self.advance()

                pos_after_brace = self.pos.copy()

                while self.current_char != "}":
                    to_lex += self.current_char
                    self.advance()

                    if self.current_char is None or self.current_char in "{\n":
                        return None, ExpectedItemError(pos_after_brace, self.pos, "Expected '}'")

                # Lex to_lex and return the error if it exists
                lexer = Lexer("fstring", to_lex)
                tokens, error = lexer.make_tokens()
                if error:
                    return None, error

                # Check if there is a string in tokens
                for token in tokens:
                    if token.type == toks.TT_STRING:
                        return None, IllegalCharError(pos_after_brace, self.pos, "Unexpected string")

                # Add the tokens to str_components
                str_components.append(tokens)
            else:
                cur_str += self.current_char

            self.advance()
            if self.current_char is None:
                return None, IllegalCharError(pos_start, self.pos, "Unexpected end to string")

        if cur_str != "":
            str_components.append(cur_str)

        self.advance()
        return Token(toks.TT_STRING, str_components, pos_start, self.pos), None

    def make_identifier(self, start_value=""):
        id_str = start_value
        pos_start = self.pos.copy()

        while self.current_char and self.current_char in constants.LETTERS_DIGITS + '_':
            id_str += self.current_char
            self.advance()

        tok_type = toks.TT_KEYWORD if id_str in toks.KEYWORDS else toks.TT_IDENTIFIER
        return Token(tok_type, id_str, pos_start, self.pos), None

    def make_not_equals(self):
        pos_start = self.pos.copy()
        self.advance()

        if self.current_char == '=':
            self.advance()
            return Token(toks.TT_NE, pos_start=pos_start, pos_end=self.pos), None

        self.advance()
        return None, ExpectedItemError(pos_start, self.pos, "Expected '=' after '!'")

    def make_equals(self):
        tok_type = toks.TT_EQ
        pos_start = self.pos.copy()
        self.advance()

        if self.current_char == '>':
            self.advance()
            tok_type = toks.TT_ARROW
        elif self.current_char == '=':
            self.advance()
            tok_type = toks.TT_EE

        return Token(tok_type, pos_start=pos_start, pos_end=self.pos)

    def make_mult_toks(self, tok_type_1, tok_type_2, switch_factor):
        tok_type = tok_type_1
        pos_start = self.pos.copy()
        self.advance()

        if self.current_char == switch_factor:
            self.advance()
            tok_type = tok_type_2

        return Token(tok_type, pos_start=pos_start, pos_end=self.pos)

    def make_plus(self):
        return(self.make_mult_toks(toks.TT_PLUS, toks.TT_PLUS_EQ, "="))

    def make_minus(self):
        return self.make_mult_toks(toks.TT_MINUS, toks.TT_MIN_EQ, "=")

    def make_mul(self):
        return(self.make_mult_toks(toks.TT_MUL, toks.TT_MUL_EQ, "="))

    def make_div(self):
        return(self.make_mult_toks(toks.TT_DIV, toks.TT_DIV_EQ, "="))

    def make_less_than(self):
        return self.make_mult_toks(toks.TT_LT, toks.TT_LTE, "=")

    def make_greater_than(self):
        return self.make_mult_toks(toks.TT_GT, toks.TT_GTE, "=")

    def skip_comment(self):
        self.advance()

        # Exclusive incase it reachs the EOF and self.current_char is None
        while self.current_char and self.current_char not in ';\n':
            self.advance()

        self.advance()



class NumberNode:
    def __init__(self, value, pos_start, pos_end):
        self.value = value

        self.pos_start = pos_start
        self.pos_end = pos_end

    def __repr__(self):
        return f'{self.value}'


class NumeralNode:
    def __init__(self, value, pos_start, pos_end):
        self.value = value

        self.pos_start = pos_start
        self.pos_end = pos_end

    def __repr__(self):
        return self.value


class StringNode:
    def __init__(self, str_components, pos_start, pos_end):
        self.str_components = str_components

        self.pos_start = pos_start
        self.pos_end = pos_end

    def __repr__(self):
        return f'{self.str_components}'


class BoolNode:
    def __init__(self, value, pos_start, pos_end):
        self.value = value

        self.pos_start = pos_start
        self.pos_end = pos_end

    def __repr__(self):
        return self.value


class ListNode:
    def __init__(self, element_nodes, pos_start, pos_end):
        self.element_nodes = element_nodes

        self.pos_start = pos_start
        self.pos_end = pos_end


class DictNode:
    def __init__(self, key_pairs, pos_start, pos_end):
        self.key_pairs = key_pairs

        self.pos_start = pos_start
        self.pos_end = pos_end


class VarAccessNode:
    def __init__(self, var_name_to_get, idxes_to_get, attrs_to_get, pos_start, pos_end):
        self.var_name_to_get = var_name_to_get
        self.idxes_to_get = idxes_to_get
        self.attrs_to_get = attrs_to_get

        self.pos_start = pos_start
        self.pos_end = pos_end


class VarAssignNode:
    def __init__(self, vars_to_set, values_to_set, assign_type, pos_start, pos_end):
        self.vars_to_set = vars_to_set
        self.values_to_set = values_to_set
        self.assign_type = assign_type

        self.pos_start = pos_start
        self.pos_end = pos_end


class BinOpNode:
    def __init__(self, left_node, op_tok, right_node, pos_start, pos_end):
        self.left_node = left_node
        self.op_tok = op_tok
        self.right_node = right_node

        self.pos_start = pos_start
        self.pos_end = pos_end

    def __repr__(self):
        return f'({self.left_node}, {self.op_tok}, {self.right_node})'


class UnaryOpNode:
    def __init__(self, op_tok, node, pos_start, pos_end):
        self.op_tok = op_tok
        self.node = node

        self.pos_start = pos_start
        self.pos_end = pos_end

    def __repr__(self):
        return f'({self.op_tok}, {self.node})'


class IfNode:
    def __init__(self, cases, else_case, pos_start, pos_end):
        self.cases = cases
        self.else_case = else_case

        self.pos_start = pos_start
        self.pos_end = pos_end


class TryNode:
    def __init__(self, try_body, except_body, except_name, except_as, should_return_null, pos_start, pos_end):
        self.try_body = try_body
        self.except_body = except_body
        self.except_name = except_name
        self.except_as = except_as
        self.should_return_null = should_return_null

        self.pos_start = pos_start
        self.pos_end = pos_end


class ForNode:
    def __init__(self, var_name, start_value_node, end_value_node, step_value_node, body_node, should_return_null, pos_start, pos_end):
        self.var_name = var_name
        self.start_value_node = start_value_node
        self.end_value_node = end_value_node
        self.step_value_node = step_value_node
        self.body_node = body_node
        self.should_return_null = should_return_null

        self.pos_start = pos_start
        self.pos_end = pos_end


class WhileNode:
    def __init__(self, condition_node, body_node, should_return_null, pos_start, pos_end):
        self.condition_node = condition_node
        self.body_node = body_node
        self.should_return_null = should_return_null

        self.pos_start = pos_start
        self.pos_end = pos_end


class FuncDefNode:
    def __init__(self, func_name, args, body_node, should_auto_return, pos_start, pos_end):
        self.func_name = func_name
        self.args = args
        self.body_node = body_node
        self.should_auto_return = should_auto_return

        self.pos_start = pos_start
        self.pos_end = pos_end


class CallNode:
    def __init__(self, node_to_call, arg_nodes, pos_start, pos_end):
        self.node_to_call = node_to_call
        self.arg_nodes = arg_nodes

        self.pos_start = pos_start
        self.pos_end = pos_end


class ReturnNode:
    def __init__(self, node_to_return, pos_start, pos_end):
        self.node_to_return = node_to_return

        self.pos_start = pos_start
        self.pos_end = pos_end


class ContinueNode:
    def __init__(self, pos_start, pos_end):
        self.pos_start = pos_start
        self.pos_end = pos_end


class BreakNode:
    def __init__(self, pos_start, pos_end):
        self.pos_start = pos_start
        self.pos_end = pos_end



#######################################
# IMPORTS
#######################################

from .Value import Value

#######################################
# BOOL
#######################################


class Bool(Value):
    def __init__(self, value):
        super().__init__()
        self.value = value

    def copy(self):
        copy = Bool(self.value)
        copy.set_pos(self.pos_start, self.pos_end)
        copy.set_context(self.context)
        return copy

    def get_comparison_eq(self, other):
        if isinstance(other, Bool):
            return Bool(self.value == other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_ne(self, other):
        if isinstance(other, Bool):
            return Bool(self.value != other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_lt(self, other):
        if isinstance(other, Bool):
            return Bool(self.value < other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_gt(self, other):
        if isinstance(other, Bool):
            return Bool(self.value > other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_lte(self, other):
        if isinstance(other, Bool):
            return Bool(self.value <= other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_gte(self, other):
        if isinstance(other, Bool):
            return Bool(self.value >= other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def anded_by(self, other):
        if isinstance(other, Bool):
            return Bool(self.value and other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def ored_by(self, other):
        if isinstance(other, Bool):
            return Bool(self.value or other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def notted(self):
        return Bool(not self.value).set_context(self.context), None

    def is_true(self):
        return self.value

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return str(self.value)



#####################################
# IMPORTS
#####################################

import os

from Celeratas.helper.errors import IndexingError, RTError, TypingError
from Celeratas.interpreter.RTResult import RTResult
from Celeratas.interpreter.values import List, Number, String

from .BaseFunction import BaseFunction

#####################################
# BUILT IN FUNCTIONS
#####################################


class BuiltInFunction(BaseFunction):
    def __init__(self, name):
        super().__init__(name)

    def execute(self, args, _):
        res = RTResult()
        exec_ctx = self.generate_new_context()

        method_name = f'execute_{self.name}'
        method = getattr(self, method_name, self.no_visit_method)

        res.register(self.check_and_populate_args(
            method.arg_names, args, exec_ctx))
        if res.should_return():
            return res

        return_value = res.register(method(exec_ctx))
        if res.should_return():
            return res
        return res.success(return_value)

    def no_visit_method(self, node, context):
        raise Exception(f'No execute_{self.name} method defined')

    def copy(self):
        copy = BuiltInFunction(self.name)
        copy.set_context(self.context)
        copy.set_pos(self.pos_start, self.pos_end)
        return copy

    def __repr__(self):
        return f"<built-in function {self.name}>"

    #####################################

    def execute_print(self, exec_ctx):
        print(str(exec_ctx.symbol_table.get('value')))
        return RTResult().success(None)
    execute_print.arg_names = [('value', None)]

    def execute_input(self, exec_ctx):
        text = input()
        return RTResult().success(String(text))
    execute_input.arg_names = []

    def execute_clear(self, exec_ctx):
        os.system('cls' if os.name == 'nt' else 'clear')
        return RTResult().success(None)
    execute_clear.arg_names = []

    def execute_is_number(self, exec_ctx):
        is_number = isinstance(exec_ctx.symbol_table.get("value"), Number)
        return RTResult().success(Number.true if is_number else Number.false)
    execute_is_number.arg_names = [('value', None)]

    def execute_is_string(self, exec_ctx):
        is_number = isinstance(exec_ctx.symbol_table.get("value"), String)
        return RTResult().success(Number.true if is_number else Number.false)
    execute_is_string.arg_names = [("value", None)]

    def execute_is_list(self, exec_ctx):
        is_number = isinstance(exec_ctx.symbol_table.get("value"), List)
        return RTResult().success(Number.true if is_number else Number.false)
    execute_is_list.arg_names = [("value", None)]

    def execute_is_function(self, exec_ctx):
        is_number = isinstance(
            exec_ctx.symbol_table.get("value"), BaseFunction)
        return RTResult().success(Number.true if is_number else Number.false)
    execute_is_function.arg_names = [("value", None)]

    def execute_append(self, exec_ctx):
        list_ = exec_ctx.symbol_table.get("list")
        value = exec_ctx.symbol_table.get("value")

        if not isinstance(list_, List):
            return RTResult().failure(TypingError(
                self.pos_start, self.pos_end,
                "First argument must be list",
                exec_ctx
            ))

        list_.elements.append(value)
        return RTResult().success(None)
    execute_append.arg_names = [("list", "value")]

    def execute_pop(self, exec_ctx):
        list_ = exec_ctx.symbol_table.get("list")
        index = exec_ctx.symbol_table.get("index")

        if not isinstance(list_, List):
            return RTResult().failure(TypingError(
                self.pos_start, self.pos_end,
                "First argument must be list",
                exec_ctx
            ))

        if not isinstance(index, Number):
            return RTResult().failure(TypingError(
                self.pos_start, self.pos_end,
                "Second argument must be number",
                exec_ctx
            ))

        try:
            element = list_.elements.pop(index.value)
        except IndexError:
            return RTResult().failure(IndexingError(
                self.pos_start, self.pos_end,
                'Element at this index could not be removed from list because index is out of bounds',
                exec_ctx
            ))
        return RTResult().success(element)
    execute_pop.arg_names = [("list", None), ("index", None)]

    def execute_extend(self, exec_ctx):
        listA = exec_ctx.symbol_table.get("listA")
        listB = exec_ctx.symbol_table.get("listB")

        if not isinstance(listA, List):
            return RTResult().failure(TypingError(
                self.pos_start, self.pos_end,
                "First argument must be list",
                exec_ctx
            ))

        if not isinstance(listB, List):
            return RTResult().failure(TypingError(
                self.pos_start, self.pos_end,
                "Second argument must be list",
                exec_ctx
            ))

        listA.elements.extend(listB.elements)
        return RTResult().success(None)
    execute_extend.arg_names = [("listA", None), ("listB", None)]

    def execute_len(self, exec_ctx):
        input_ = exec_ctx.symbol_table.get("input")

        if not isinstance(input_, List) and not isinstance(input_, String):
            return RTResult().failure(TypingError(
                self.pos_start, self.pos_end,
                "Argument must be list or string",
                exec_ctx
            ))

        return RTResult().success(Number(len(input_.elements if isinstance(input_, List) else input_.value)))
    execute_len.arg_names = [("input", None)]

    def execute_run(self, exec_ctx):
        fn = exec_ctx.symbol_table.get("fn")

        if not isinstance(fn, String):
            return RTResult().failure(TypingError(
                self.pos_start, self.pos_end,
                "Second argument must be string",
                exec_ctx
            ))

        fn = fn.value

        try:
            with open(fn, "r") as f:
                script = f.read()
        except Exception:
            return RTResult().failure(RTError(
                self.pos_start, self.pos_end,
                f"Failed to load script \"{fn}\"",
                exec_ctx
            ))

        # Need the import here to avoid circular import
        from Celeratas.shell import run_script
        _, error = run_script(fn, script)

        if error:
            return RTResult().failure(TypingError(
                self.pos_start, self.pos_end,
                f"Failed to finish executing script \"{fn}\"\n" + error.as_string(),
                exec_ctx
            ))

        return RTResult().success(None)
    execute_run.arg_names = [("fn", None)]



#####################################
# IMPORTS
#####################################

from Celeratas.interpreter.RTResult import RTResult

from .BaseFunction import BaseFunction


#####################################
# FUNCTION
#####################################
class Function(BaseFunction):
    def __init__(self, name, body_node, arg_names, should_auto_return):
        super().__init__(name)
        self.body_node = body_node
        self.arg_names = arg_names
        self.should_auto_return = should_auto_return

    def execute(self, args, recursion_depth):
        res = RTResult()
        from Celeratas.interpreter.Interpreter import Interpreter
        interpreter = Interpreter(recursion_depth + 1)
        exec_ctx = self.generate_new_context()

        res.register(self.check_and_populate_args(
            self.arg_names, args, exec_ctx))
        if res.should_return():
            return res

        value = res.register(interpreter.visit(self.body_node, exec_ctx))
        if res.should_return() and res.func_return_value is None:
            return res

        ret_value = (
            value if self.should_auto_return else None) or res.func_return_value or None
        return res.success(ret_value)

    def copy(self):
        copy = Function(self.name, self.body_node,
                        self.arg_names, self.should_auto_return)
        copy.set_context(self.context)
        copy.set_pos(self.pos_start, self.pos_end)
        return copy

    def __repr__(self):
        return f"<function {self.name}>"



#######################################
# IMPORTS
#######################################

from .Value import Value

#######################################
# LIST
#######################################


class List(Value):
    def __init__(self, elements):
        super().__init__()
        self.elements = elements

    def copy(self):
        copy = List(self.elements)
        copy.set_pos(self.pos_start, self.pos_end)
        copy.set_context(self.context)
        return copy

    def __str__(self):
        return ", ".join([str(x) for x in self.elements])

    def __repr__(self):
        return f'[{", ".join([repr(x) for x in self.elements])}]'



#######################################
# IMPORTS
#######################################

from .Value import Value

#######################################
# DICT
#######################################


class Dict(Value):
    def __init__(self, key_pairs):
        super().__init__()
        self.key_pairs = key_pairs

    def copy(self):
        copy = Dict(self.key_pairs)
        copy.set_pos(self.pos_start, self.pos_end)
        copy.set_context(self.set_context)
        return copy

    def __str__(self):
        return str(self.key_pairs)

    def __repr__(self):
        return str(self.key_pairs)
